import itertools
import math
import textwrap

import torch

from ninetoothed.eval import _generate_target_tensor_shape
from ninetoothed.generation import cache_source
from ninetoothed.jit import import_from_path
from ninetoothed.make import make
from ninetoothed.tensor import Tensor


def simulate_arrangement(arrangement, tensors, device=None):
    """Simulate the arrangement of the tensors.

    :param arrangement: The arrangement of the tensors.
    :param tensors: The tensors.
    :param device: The device on which the tensors are.
    :return: A tuple of source tensors and a tuple of target tensors,
        where each tuple is sorted according to the parameter order
        specified in ``arrangement``, and each element in each tensor
        stores the index of that element in the source tensor.
    """

    if device is None:
        device = "cuda"

    def _arrangement(*tensors):
        return tensors

    source_tensors = []
    target_tensors = []

    for tensor, arranged in zip(tensors, arrangement(*tensors)):
        if tensor.ndim == 0:
            source_tensors.append(torch.tensor(0, device=device))
            target_tensors.append(torch.tensor(0, device=device))

            continue

        num_programs = math.prod(tensor.shape)

        shape = _generate_target_tensor_shape(arranged, flatten_outermost=True)

        source_tensor = torch.arange(num_programs, device=device).view(tensor.shape)
        target_tensor = torch.empty(shape, dtype=source_tensor.dtype, device=device)

        source_tensors.append(source_tensor)
        target_tensors.append(target_tensor)

    tensors = arrangement(
        *(
            Tensor(tensor.ndim, other=-1, shape_options={"constexpr": True})
            for tensor in tensors
        )
    )
    debug_tensors = tuple(_generate_debug_tensor(tensor) for tensor in tensors)

    application_source = _generate_debug_application_source(tensors, debug_tensors)

    source_file = str(cache_source(application_source))

    module = import_from_path(source_file, source_file)
    module_vars = vars(module)

    application = module_vars[_APPLICATION_NAME]

    kernel = make(_arrangement, application, tensors + debug_tensors)

    kernel(*source_tensors, *target_tensors)

    return tuple(source_tensors), tuple(target_tensors)


_INDENT = "    "

_APPLICATION_NAME = "application"


def _generate_debug_application_source(tensors, debug_tensors):
    params = ", ".join(
        tensor.source.name for tensor in itertools.chain(tensors, debug_tensors)
    )

    body_lines = []

    for tensor, debug_tensor in zip(tensors, debug_tensors):
        assignment_lines = _generate_debug_assignment_lines(tensor, debug_tensor)

        body_lines.extend(assignment_lines)

    body = textwrap.indent("\n".join(body_lines), _INDENT)

    return f"def {_APPLICATION_NAME}({params}):\n{body}"


def _generate_debug_assignment_lines(tensor, debug_tensor):
    if tensor.ndim == 0:
        return []

    num_indices = 0

    curr = debug_tensor.dtype

    while isinstance(curr.dtype, type(debug_tensor)):
        num_indices += curr.ndim

        curr = curr.dtype

    lines = []

    indices = []

    for i in range(num_indices):
        index = chr(ord("i") + i)
        line = textwrap.indent(
            f"for {index} in range({debug_tensor.source.name}.shape[{i}]):", _INDENT * i
        )

        lines.append(line)
        indices.append(index)

    if indices:
        joined_indices = ", ".join(indices)

        lines.append(
            textwrap.indent(
                f"{debug_tensor.source.name}[{joined_indices}] = {tensor.source.name}[{joined_indices}]",
                _INDENT * num_indices,
            )
        )
    else:
        lines.append(
            textwrap.indent(
                f"{debug_tensor.source.name} = {tensor.source.name}",
                _INDENT * num_indices,
            )
        )

    return lines


def _generate_debug_tensor(tensor):
    if tensor.ndim == 0:
        return Tensor(tensor.ndim)

    levels = []

    curr = tensor

    while isinstance(curr, type(tensor)):
        levels.append(curr)

        curr = curr.dtype

    ndim = sum(level.ndim for level in levels[1:]) + 1

    tile_shapes = []

    start = 0

    for level in reversed(levels[1:]):
        stop = start + level.ndim

        tile_shape = tuple(-1 if start <= i < stop else 1 for i in range(ndim))
        tile_shape = tuple(reversed(tile_shape))

        tile_shapes.append(tile_shape)

        start = stop

    debug_tensor = Tensor(ndim, shape_options={"constexpr": True})

    for tile_shape in tile_shapes:
        singleton_dims = tuple(i for i in range(ndim) if tile_shape[i] == 1)

        debug_tensor = debug_tensor.tile(tile_shape)
        debug_tensor.dtype = debug_tensor.dtype.squeeze(singleton_dims)

    return debug_tensor



import torch

import ninetoothed
from ninetoothed import Tensor, block_size
from ninetoothed.language import libdevice
from tests.skippers import skip_if_cuda_not_available


def arrangement(input, exponent, output, BLOCK_SIZE=block_size()):
    return (
        input.tile((BLOCK_SIZE,)),
        exponent.tile((BLOCK_SIZE,)),
        output.tile((BLOCK_SIZE,)),
    )


def application(input, exponent, output):
    output = libdevice.pow(input, exponent)  # noqa: F841


def pow(input, exponent):
    output = torch.empty_like(input)

    pow_kernel = ninetoothed.make(
        arrangement, application, (Tensor(1), Tensor(1), Tensor(1))
    )

    pow_kernel(input, exponent, output)

    return output


@skip_if_cuda_not_available
class TestCUDA:
    @classmethod
    def setup_class(cls):
        torch.manual_seed(0)

        size = 44925

        cls.input = torch.randn(size, device="cuda")
        cls.exponent = torch.randn(size, device="cuda")

    def test_fp32(self):
        input = type(self).input.to(torch.float32)
        exponent = type(self).exponent.to(torch.float32)

        assert torch.allclose(
            pow(input, exponent), torch.pow(input, exponent), equal_nan=True
        )



import ctypes
import functools
import subprocess

import torch
import torch.nn.functional as F

import ninetoothed
import ninetoothed.generation
import tests.test_addmm as addmm
import tests.test_attention as attention
import tests.test_conv2d as conv2d
import tests.test_matmul as matmul
from ninetoothed import Tensor
from tests.skippers import skip_if_cuda_not_available


@skip_if_cuda_not_available
class TestCUDA:
    @classmethod
    def setup_class(cls):
        torch.manual_seed(0)

    def test_add(self):
        def _arrangement(input, other, output):
            def _arrange(tensor):
                return tensor.tile((256,))

            return _arrange(input), _arrange(other), _arrange(output)

        def _application(input, other, output):
            output = input + other  # noqa: F841

        tensors = tuple(Tensor(1, dtype=ninetoothed.bfloat16) for _ in range(3))
        caller = "cuda"
        kernel_name = "add"
        output_dir = ninetoothed.generation.CACHE_DIR

        launch_func = _generate_launch_func(
            _arrangement,
            _application,
            tensors,
            caller=caller,
            kernel_name=kernel_name,
            output_dir=output_dir,
        )

        shape = (45327,)
        dtype = torch.bfloat16
        device = caller

        input = torch.randn(shape, dtype=dtype, device=device)
        other = torch.randn(shape, dtype=dtype, device=device)
        output = torch.empty_like(input)

        _run_launch_func(launch_func, input, other, output)

        assert torch.allclose(output, torch.add(input, other))

    def test_addmm(self):
        arrangement = functools.partial(
            addmm.arrangement, BLOCK_SIZE_M=64, BLOCK_SIZE_N=64, BLOCK_SIZE_K=64
        )
        application = addmm.application
        tensors = tuple(
            Tensor(ndim, dtype=ninetoothed.float16) for ndim in (2, 2, 2, 0, 0, 2)
        )
        caller = "cuda"
        kernel_name = "addmm"
        output_dir = ninetoothed.generation.CACHE_DIR

        launch_func = _generate_launch_func(
            arrangement,
            application,
            tensors,
            caller=caller,
            kernel_name=kernel_name,
            output_dir=output_dir,
        )

        shape = (512, 512)
        dtype = torch.float16
        device = caller

        input = torch.randn(shape, dtype=dtype, device=device)
        mat1 = torch.randn(shape, dtype=dtype, device=device)
        mat2 = torch.randn(shape, dtype=dtype, device=device)
        beta = torch.randn((), dtype=dtype)
        alpha = torch.randn((), dtype=dtype)
        output = torch.empty(
            (mat1.shape[0], mat2.shape[1]), dtype=mat1.dtype, device=mat1.device
        )

        _run_launch_func(launch_func, input, mat1, mat2, beta, alpha, output)

        assert torch.allclose(
            output, torch.addmm(input, mat1, mat2, beta=beta, alpha=alpha), atol=0.075
        )

    def test_attention(self):
        emb_dim = 64

        arrangement = functools.partial(
            attention.arrangement, BLOCK_SIZE_M=64, BLOCK_SIZE_N=64
        )
        application = attention.application
        query_, key_, value_, output_ = tuple(
            Tensor(4, dtype=ninetoothed.float16) for _ in range(4)
        )
        for tensor in (query_, key_, value_, output_):
            tensor.shape = tensor.shape[:-1] + (emb_dim,)
        is_causal_ = Tensor(0, constexpr=True, value=1)
        tensors = (query_, key_, value_, is_causal_, output_)
        caller = "cuda"
        kernel_name = "attention"
        output_dir = ninetoothed.generation.CACHE_DIR

        launch_func = _generate_launch_func(
            arrangement,
            application,
            tensors,
            caller=caller,
            kernel_name=kernel_name,
            output_dir=output_dir,
        )

        shape = (2, 4, 1024, emb_dim)
        dtype = torch.float16
        device = caller

        query = torch.randn(shape, dtype=dtype, device=device)
        key = torch.randn(shape, dtype=dtype, device=device)
        value = torch.randn(shape, dtype=dtype, device=device)
        is_causal = torch.tensor(True)
        output = torch.empty(shape, dtype=dtype, device=device)

        _run_launch_func(launch_func, query, key, value, is_causal, output)

        assert torch.allclose(
            output,
            F.scaled_dot_product_attention(query, key, value, is_causal=True, scale=1),
            atol=0.01,
        )

    def test_matmul(self):
        arrangement = functools.partial(
            matmul.arrangement, BLOCK_SIZE_M=64, BLOCK_SIZE_N=64, BLOCK_SIZE_K=64
        )
        application = matmul.application
        tensors = tuple(Tensor(2, dtype=ninetoothed.float16) for _ in range(3))
        caller = "cuda"
        kernel_name = "matmul"
        output_dir = ninetoothed.generation.CACHE_DIR

        launch_func = _generate_launch_func(
            arrangement,
            application,
            tensors,
            caller=caller,
            kernel_name=kernel_name,
            output_dir=output_dir,
        )

        shape = (512, 512)
        dtype = torch.float16
        device = caller

        lhs = torch.randn(shape, dtype=dtype, device=device)
        rhs = torch.randn(shape, dtype=dtype, device=device)
        output = torch.empty((lhs.shape[0], rhs.shape[1]), dtype=dtype, device=device)

        _run_launch_func(launch_func, lhs, rhs, output)

        assert torch.allclose(output, torch.matmul(lhs, rhs))

    def test_conv2d(self):
        arrangement = functools.partial(
            conv2d.arrangement, BLOCK_SIZE_M=64, BLOCK_SIZE_N=64, BLOCK_SIZE_K=64
        )
        application = matmul.application
        tensors = tuple(Tensor(4, dtype=ninetoothed.float16) for _ in range(3))
        caller = "cuda"
        kernel_name = "conv2d"
        output_dir = ninetoothed.generation.CACHE_DIR

        launch_func = _generate_launch_func(
            arrangement,
            application,
            tensors,
            caller=caller,
            kernel_name=kernel_name,
            output_dir=output_dir,
        )

        n, c, h, w = 4, 64, 16, 16
        k, _, r, s = 512, c, 3, 3
        p = h - r + 1
        q = w - s + 1
        dtype = torch.float16
        device = caller

        input = torch.randn(n, c, h, w, dtype=dtype, device=device)
        filter = torch.randn(k, c, r, s, dtype=dtype, device=device)
        output = torch.empty(n, k, p, q, dtype=dtype, device=device)

        _run_launch_func(launch_func, input, filter, output)

        assert torch.allclose(output, F.conv2d(input, filter), atol=0.001, rtol=0.001)


class _ArgumentTensor(ctypes.Structure):
    _fields_ = [
        ("data", ctypes.c_void_p),
        ("shape", ctypes.POINTER(ctypes.c_uint64)),
        ("strides", ctypes.POINTER(ctypes.c_int64)),
    ]

    @staticmethod
    def from_torch_tensor(tensor):
        data = ctypes.c_void_p(tensor.data_ptr())
        shape = (ctypes.c_uint64 * len(tensor.shape))(*tensor.shape)
        strides = (ctypes.c_int64 * len(tensor.stride()))(*tensor.stride())

        return _ArgumentTensor(data, shape, strides)


def _run_launch_func(launch_func, *tensors):
    stream = torch.cuda.Stream()

    arg_tensors = tuple(_ArgumentTensor.from_torch_tensor(tensor) for tensor in tensors)

    with torch.cuda.stream(stream):
        launch_func(ctypes.c_void_p(stream.cuda_stream), *arg_tensors)

    stream.synchronize()


def _generate_launch_func(
    arrangement, application, tensors, caller, kernel_name, output_dir
):
    ninetoothed.make(
        arrangement,
        application,
        tensors,
        caller=caller,
        kernel_name=kernel_name,
        output_dir=output_dir,
    )

    _compile_library(kernel_name, output_dir)
    library = _load_library(kernel_name, output_dir)
    launch_func_name = f"launch_{kernel_name}"
    launch_func = getattr(library, launch_func_name)
    launch_func.argtypes = (ctypes.c_void_p,) + tuple(_ArgumentTensor for _ in tensors)
    launch_func.restype = ctypes.c_int

    return launch_func


def _compile_library(kernel_name, output_dir):
    command = [
        "nvcc",
        "-shared",
        "-Xcompiler",
        "-fPIC",
        "-lcuda",
        "-o",
        output_dir / f"{kernel_name}.so",
        output_dir / f"{kernel_name}.c",
    ]

    subprocess.run(command, check=True)


def _load_library(kernel_name, kernel_dir):
    return ctypes.CDLL(kernel_dir / f"{kernel_name}.so")



import copy
import functools
import itertools
import math
import re

import ninetoothed.naming as naming
from ninetoothed.symbol import Symbol


class Tensor:
    """A class uesed to represent a symbolic tensor.

    :param ndim: The number of dimensions of the tensor.
    :param shape: The shape of the tensor.
    :param dtype: The element type of the tensor.
    :param other: The values for out-of-bounds positions.
    :param shape_options: The options for configuring shape symbols.
    :param name: The name of the tensor.
    :param source: For internal use only.
    :param target_dims: For internal use only.
    """

    num_instances = 0

    def __init__(
        self,
        ndim=None,
        shape=None,
        dtype=None,
        other=None,
        shape_options=None,
        constexpr=None,
        value=None,
        name=None,
        source=None,
        target_dims=None,
        _offsets=None,
        _outputs=None,
    ):
        self.dtype = dtype

        if name is not None:
            self.name = name
        else:
            self.name = naming.auto_generate(f"tensor_{type(self).num_instances}")

        if ndim is not None:
            if shape_options is None:
                shape_options = tuple({} for _ in range(ndim))

            if isinstance(shape_options, dict):
                shape_options = tuple(shape_options for _ in range(ndim))

            shape_options = tuple(
                size_options if size_options is not None else {}
                for size_options in shape_options
            )

            self.shape = (
                Symbol(self.size_string(i), **size_options)
                for i, size_options in zip(range(ndim), shape_options)
            )
        else:
            self.shape = shape

        self.other = other

        if constexpr and self.ndim != 0:
            raise ValueError(
                "`constexpr` can only be set for zero-dimensional tensors."
            )

        self.constexpr = constexpr

        if self.constexpr:
            self.name = naming.make_constexpr(self.name)

        if not constexpr and value is not None:
            raise ValueError("`value` can only be set for constexpr tensors.")

        self.value = value

        if source is not None:
            self.source = source
        else:
            self.source = self

        if target_dims is not None:
            self.target_dims = target_dims
        else:
            self.target_dims = (dim for dim in range(self.ndim))

        if _offsets is not None:
            self._levels = self.source._levels

            self._offsets = _offsets

            self._outputs = _outputs
        else:
            self._levels = [[self]]

            def _offsets(indices):
                return (tuple(indices),)

            self._offsets = _offsets

            self._outputs = [[]]

        self._inputs = []

        self._history = []

        type(self).num_instances += 1

    @staticmethod
    def _meta_operation(func):
        @functools.wraps(func)
        def wrapper(self, *args, **kwargs):
            if self.source is self:
                self = copy.deepcopy(self)

            result = func(self, *args, **kwargs)

            result._history.extend(self._history)
            result._history.append((func, args, kwargs))

            return result

        return wrapper

    @_meta_operation
    def tile(self, tile_shape, strides=None, dilation=None, floor_mode=False):
        """Tiles the tensor into a hierarchical tensor.

        :param tile_shape: The shape of a tile.
        :param strides: The interval at which each tile is generated.
        :param dilation: The spacing between tiles.
        :param floor_mode: If ``True``, will use floor division to
            compute the outer shape.
        :return: A hierarchical tensor.
        """

        if strides is None:
            strides = [-1 for _ in tile_shape]

        if dilation is None:
            dilation = [1 for _ in tile_shape]

        outer_shape = []
        outer_strides = []
        inner_shape = []
        inner_strides = []

        for self_size, tile_size, stride, spacing in zip(
            self.shape, tile_shape, strides, dilation
        ):
            if tile_size == -1:
                tile_size = self_size

            if stride == -1:
                stride = tile_size

            def _div(x, y, floor_mode=False):
                if floor_mode:
                    return x // y

                return (x + y - 1) // y

            new_size = (
                (
                    _div(
                        self_size - spacing * (tile_size - 1) - 1,
                        stride,
                        floor_mode=floor_mode,
                    )
                    + 1
                )
                if stride != 0
                else -1
            )

            outer_shape.append(new_size)
            outer_strides.append(stride)

            inner_shape.append(tile_size)
            inner_strides.append(spacing)

        self._inputs.extend(([], []))

        def _offsets(indices):
            return (
                tuple(index * stride for index, stride in zip(indices, inner_strides)),
            )

        dtype = type(self)(
            shape=inner_shape,
            dtype=self.dtype,
            source=self.source,
            _offsets=_offsets,
            _outputs=[self._inputs[1]],
        )

        def _offsets(indices):
            return (
                tuple(index * stride for index, stride in zip(indices, outer_strides)),
            )

        output = type(self)(
            shape=outer_shape,
            dtype=dtype,
            source=self.source,
            _offsets=_offsets,
            _outputs=[self._inputs[0]],
        )

        self._levels.append([output, dtype])

        return output

    @_meta_operation
    def expand(self, shape):
        """Expands the specified singleton dimensions of the tensor.

        :param shape: The expanded shape.
        :return: The expanded tensor.
        """

        self._inputs.append([])

        def _offsets(indices):
            return (
                tuple(
                    index if new_size == -1 else 0
                    for index, new_size in zip(indices, shape)
                ),
            )

        # TODO: Add error handling.
        output = type(self)(
            shape=[
                new_size if new_size != -1 else size
                for size, new_size in zip(self.shape, shape)
            ],
            dtype=self.dtype,
            source=self.source,
            target_dims=self.target_dims,
            _offsets=_offsets,
            _outputs=[self._inputs[0]],
        )

        self._levels.append([output])

        return output

    @_meta_operation
    def squeeze(self, dim):
        """Removes the specified singleton dimensions of the tensor.

        :param dim: The dimension(s) to be squeezed.
        :return: The squeezed tensor.
        """

        if not isinstance(dim, tuple):
            dim = (dim,)

        self._inputs.append([])

        def _offsets(indices):
            return (
                (
                    lambda iter: tuple(
                        next(iter) if i not in dim else 0 for i in range(self.ndim)
                    )
                )(iter(indices)),
            )

        # TODO: Add error handling.
        output = type(self)(
            shape=[size for i, size in enumerate(self.shape) if i not in dim],
            dtype=self.dtype,
            source=self.source,
            target_dims=[
                target_dim
                for i, target_dim in enumerate(self.target_dims)
                if i not in dim
            ],
            _offsets=_offsets,
            _outputs=[self._inputs[0]],
        )

        self._levels.append([output])

        return output

    @_meta_operation
    def permute(self, dims):
        """Permutes the dimensions of the tensor.

        :param dims: The permuted ordering of the dimensions.
        :return: The permuted tensor.
        """

        # TODO: Add error handling.
        new_shape = [None for _ in range(self.ndim)]

        for original_dim, permuted_dim in enumerate(dims):
            new_shape[original_dim] = self.shape[permuted_dim]

        self._inputs.append([])

        def _offsets(indices):
            return (tuple(indices[dims.index(dim)] for dim in range(len(dims))),)

        output = type(self)(
            shape=new_shape,
            dtype=self.dtype,
            source=self.source,
            target_dims=self.target_dims,
            _offsets=_offsets,
            _outputs=[self._inputs[0]],
        )

        self._levels.append([output])

        return output

    @_meta_operation
    def flatten(self, start_dim=None, end_dim=None):
        """Flattens the specified dimensions of the tensor.

        See :func:`ravel` for the differences between :func:`flatten`
        and :func:`ravel`.

        :param start_dim: The first dimension to flatten.
        :param end_dim: The dimension after the last to flatten.
        :return: The flattened tensor.
        """

        # TODO: Add error handling.
        if start_dim is None:
            start_dim = 0
        if end_dim is None:
            end_dim = self.ndim

        leading_sizes = self.shape[:start_dim]
        flattening_sizes = self.shape[start_dim:end_dim]
        trailing_sizes = self.shape[end_dim:]

        new_shape = leading_sizes + (math.prod(flattening_sizes),) + trailing_sizes

        leading_target_dims = self.target_dims[:start_dim]
        flattening_target_dims = self.target_dims[start_dim:end_dim]
        trailing_target_dims = self.target_dims[end_dim:]

        new_target_dims = (
            leading_target_dims + (flattening_target_dims[-1],) + trailing_target_dims
        )

        self._inputs.append([])

        def _offsets(indices):
            start_dim_ = self.ndim + start_dim if start_dim < 0 else start_dim

            return (
                (
                    indices[:start_dim_]
                    + type(self)._unravel_index(indices[start_dim_], flattening_sizes)
                    + indices[start_dim_ + 1 :]
                ),
            )

        output = type(self)(
            shape=new_shape,
            dtype=self.dtype,
            source=self.source,
            target_dims=new_target_dims,
            _offsets=_offsets,
            _outputs=[self._inputs[0]],
        )

        self._levels.append([output])

        return output

    @_meta_operation
    def ravel(self):
        """Flattens the hierarchy of the tensor.

        :func:`ravel` differs from :func:`flatten`, which only flattens
        dimensions at a single level. For example, consider a tensor
        with two levels: the first level has a shape of ``(N, P, Q)``,
        and the second level has a shape of ``(C, R, S)``. After
        applying :func:`ravel`, the resulting tensor will have a single
        flattened level with a shape of ``(N, P, Q, C, R, S)``.

        :return: The raveled tensor.
        """

        # TODO: Add error handling.
        new_shape = []
        outputs = []

        curr = self

        while isinstance(curr, type(self)):
            new_shape.extend(curr.shape)
            curr._inputs.append([])
            outputs.extend(curr._inputs)

            curr = curr.dtype

        def _offsets(indices):
            outputs = []

            curr = self
            start = 0

            while isinstance(curr, type(self)):
                stop = start + curr.ndim
                curr_indices = indices[start:stop]

                outputs.append(curr_indices)

                start = stop
                curr = curr.dtype

            return tuple(outputs)

        output = type(self)(
            shape=new_shape,
            other=self.source.other,
            name=self.source.name,
            source=self.source,
            _offsets=_offsets,
            _outputs=outputs,
        )

        self._levels.append([output])

        return output

    def offsets(self):
        for output_, output in zip(
            self._outputs,
            self._offsets(tuple(sum(indices) for indices in zip(*self._inputs))),
        ):
            output_.clear()
            output_.extend(output)

    def names(self):
        if self.ndim == 0:
            return {Symbol(self.source.name)}

        strides = tuple(
            Symbol(self.source.stride_string(dim)) for dim in range(self.ndim)
        )

        return (
            {Symbol(self.source.pointer_string())}
            | {
                name
                for value in itertools.chain(self.shape, strides)
                if isinstance(value, Symbol)
                for name in value.names()
            }
            | (self.dtype.names() if isinstance(self.dtype, type(self)) else set())
            | (self.source.names() if self.source is not self else set())
        )

    def innermost(self):
        if not isinstance(self.dtype, type(self)):
            return self

        return self.dtype.innermost()

    def pointer_string(self):
        return naming.auto_generate(f"{self.name}_pointer")

    def size_string(self, dim):
        return naming.auto_generate(f"{self.name}_size_{dim}")

    def stride_string(self, dim):
        return naming.auto_generate(f"{self.name}_stride_{dim}")

    def size(self, dim=None):
        if dim is None:
            return self.shape

        return self.shape[dim]

    @property
    def shape(self):
        return self._shape

    @shape.setter
    def shape(self, value):
        self._shape = tuple(value)

    @property
    def ndim(self):
        return len(self.shape)

    @property
    def target_dims(self):
        return self._target_dims

    @target_dims.setter
    def target_dims(self, value):
        self._target_dims = tuple(value)

    @staticmethod
    def pointer_pattern():
        return re.compile(
            naming.auto_generate(rf"({_identifier_pattern_raw_string()})_(pointer)")
        )

    @staticmethod
    def size_pattern():
        return re.compile(
            naming.auto_generate(rf"({_identifier_pattern_raw_string()})_(size)_(.+)")
        )

    @staticmethod
    def stride_pattern():
        return re.compile(
            naming.auto_generate(rf"({_identifier_pattern_raw_string()})_(stride)_(.+)")
        )

    @staticmethod
    def _unravel_index(index, shape):
        indices = []

        for stride in Tensor._calculate_default_strides(shape):
            indices.append(index // stride)
            index %= stride

        return tuple(indices)

    @staticmethod
    def _calculate_default_strides(shape):
        strides = [1]

        for size in reversed(shape[1:]):
            strides.append(size * strides[-1])

        return reversed(strides)


def _identifier_pattern_raw_string():
    return r"[a-zA-Z_][a-zA-Z0-9_]*"



import functools
import itertools

import torch

import tests.test_addmm as addmm
from ninetoothed import Tensor
from ninetoothed.debugging import simulate_arrangement
from tests.skippers import skip_if_cuda_not_available


@skip_if_cuda_not_available
def test_addmm():
    m = 3
    n = 4
    k = 5

    block_size_m = 2
    block_size_n = 2
    block_size_k = 2

    arrangement = functools.partial(
        addmm.arrangement,
        BLOCK_SIZE_M=block_size_m,
        BLOCK_SIZE_N=block_size_n,
        BLOCK_SIZE_K=block_size_k,
    )
    tensors = (
        Tensor(shape=(m, n)),
        Tensor(shape=(m, k)),
        Tensor(shape=(k, n)),
        Tensor(0),
        Tensor(0),
        Tensor(shape=(m, n)),
    )

    reference_source_tensors = (
        torch.tensor(
            [
                [0, 1, 2, 3],
                [4, 5, 6, 7],
                [8, 9, 10, 11],
            ]
        ),
        torch.tensor(
            [
                [0, 1, 2, 3, 4],
                [5, 6, 7, 8, 9],
                [10, 11, 12, 13, 14],
            ]
        ),
        torch.tensor(
            [
                [0, 1, 2, 3],
                [4, 5, 6, 7],
                [8, 9, 10, 11],
                [12, 13, 14, 15],
                [16, 17, 18, 19],
            ]
        ),
        torch.tensor(0),
        torch.tensor(0),
        torch.tensor(
            [
                [0, 1, 2, 3],
                [4, 5, 6, 7],
                [8, 9, 10, 11],
            ]
        ),
    )

    reference_target_tensors = (
        torch.tensor(
            [
                [
                    [0, 1],
                    [4, 5],
                ],
                [
                    [2, 3],
                    [6, 7],
                ],
                [
                    [8, 9],
                    [-1, -1],
                ],
                [
                    [10, 11],
                    [-1, -1],
                ],
            ]
        ),
        torch.tensor(
            [
                [
                    [
                        [0, 1],
                        [5, 6],
                    ],
                    [
                        [2, 3],
                        [7, 8],
                    ],
                    [
                        [4, -1],
                        [9, -1],
                    ],
                ],
                [
                    [
                        [0, 1],
                        [5, 6],
                    ],
                    [
                        [2, 3],
                        [7, 8],
                    ],
                    [
                        [4, -1],
                        [9, -1],
                    ],
                ],
                [
                    [
                        [10, 11],
                        [-1, -1],
                    ],
                    [
                        [12, 13],
                        [-1, -1],
                    ],
                    [
                        [14, -1],
                        [-1, -1],
                    ],
                ],
                [
                    [
                        [10, 11],
                        [-1, -1],
                    ],
                    [
                        [12, 13],
                        [-1, -1],
                    ],
                    [
                        [14, -1],
                        [-1, -1],
                    ],
                ],
            ]
        ),
        torch.tensor(
            [
                [
                    [
                        [0, 1],
                        [4, 5],
                    ],
                    [
                        [8, 9],
                        [12, 13],
                    ],
                    [
                        [16, 17],
                        [-1, -1],
                    ],
                ],
                [
                    [
                        [2, 3],
                        [6, 7],
                    ],
                    [
                        [10, 11],
                        [14, 15],
                    ],
                    [
                        [18, 19],
                        [-1, -1],
                    ],
                ],
                [
                    [
                        [0, 1],
                        [4, 5],
                    ],
                    [
                        [8, 9],
                        [12, 13],
                    ],
                    [
                        [16, 17],
                        [-1, -1],
                    ],
                ],
                [
                    [
                        [2, 3],
                        [6, 7],
                    ],
                    [
                        [10, 11],
                        [14, 15],
                    ],
                    [
                        [18, 19],
                        [-1, -1],
                    ],
                ],
            ]
        ),
        torch.tensor(0),
        torch.tensor(0),
        torch.tensor(
            [
                [
                    [0, 1],
                    [4, 5],
                ],
                [
                    [2, 3],
                    [6, 7],
                ],
                [
                    [8, 9],
                    [-1, -1],
                ],
                [
                    [10, 11],
                    [-1, -1],
                ],
            ]
        ),
    )

    source_tensors, target_tensors = simulate_arrangement(arrangement, tensors)

    for tensor, reference_tensor in itertools.chain(
        zip(source_tensors, reference_source_tensors),
        zip(target_tensors, reference_target_tensors),
    ):
        assert torch.equal(tensor.to(reference_tensor.device), reference_tensor)

    for tensor, reference_tensor in zip(
        (
            arranged.flatten().eval() if arranged.ndim != 0 else arranged.eval()
            for arranged in arrangement(*tensors)
        ),
        reference_target_tensors,
    ):
        assert torch.equal(torch.from_numpy(tensor), reference_tensor)



from tests.skippers import skip_if_cuda_not_available


@skip_if_cuda_not_available
def test_ipynb():
    import inspect
    import os
    import subprocess
    import textwrap

    def _test_sin():
        import torch

        import ninetoothed
        from ninetoothed import Tensor, block_size
        from ninetoothed.language import libdevice

        def arrangement(input, output, BLOCK_SIZE=block_size()):
            return (input.tile((BLOCK_SIZE,)), output.tile((BLOCK_SIZE,)))

        def application(input, output):
            output = libdevice.sin(input)  # noqa: F841

        def sin(input):
            output = torch.empty_like(input)

            sin_kernel = ninetoothed.make(
                arrangement, application, (Tensor(1), Tensor(1))
            )

            sin_kernel(input, output)

            return output

        size = 51252
        device = "cuda"

        input = torch.randn(size, device=device)

        assert torch.allclose(sin(input), torch.sin(input))

    PARENT_DIR_PATH = os.path.dirname(__file__)
    PY_FILE_NAME = "test_sin.py"
    PY_FILE_PATH = os.path.join(PARENT_DIR_PATH, PY_FILE_NAME)
    IPYNB_FILE_NAME = os.path.splitext(PY_FILE_NAME)[0] + ".ipynb"
    IPYNB_FILE_PATH = os.path.join(PARENT_DIR_PATH, IPYNB_FILE_NAME)
    NBCONVERT_IPYNB_FILE_NAME = os.path.splitext(PY_FILE_NAME)[0] + ".nbconvert.ipynb"
    NBCONVERT_IPYNB_FILE_PATH = os.path.join(PARENT_DIR_PATH, NBCONVERT_IPYNB_FILE_NAME)

    with open(PY_FILE_PATH, "w") as f:
        f.write(textwrap.dedent("".join(inspect.getsourcelines(_test_sin)[0][1:])))

    subprocess.run(("jupytext", "--to", "ipynb", PY_FILE_PATH), check=True)
    subprocess.run(
        ("jupyter", "nbconvert", "--to", "notebook", "--execute", IPYNB_FILE_PATH),
        check=True,
    )

    os.remove(PY_FILE_PATH)
    os.remove(IPYNB_FILE_PATH)
    os.remove(NBCONVERT_IPYNB_FILE_PATH)



import matplotlib.pyplot as plt

from ninetoothed import Tensor
from ninetoothed.visualization import (
    _prepare_figure_and_axes,
    _visualize_tensor,
    visualize,
)


def visualize_x_tiled(m, n, block_size_m, block_size_n):
    BLOCK_SIZE_M = block_size_m
    BLOCK_SIZE_N = block_size_n

    x = Tensor(shape=(m, n))
    x_tiled = x.tile((BLOCK_SIZE_M, BLOCK_SIZE_N))

    visualize(x_tiled, color="C0", save_path="generated/x-tiled.png")


def visualize_add(size, block_size):
    BLOCK_SIZE = block_size

    x = Tensor(shape=(size,))
    y = Tensor(shape=(size,))
    z = Tensor(shape=(size,))

    x_arranged = x.tile((BLOCK_SIZE,))
    visualize(x_arranged, color="C0", save_path="generated/x-arranged.png")

    y_arranged = y.tile((BLOCK_SIZE,))
    visualize(y_arranged, color="C1", save_path="generated/y-arranged.png")

    z_arranged = z.tile((BLOCK_SIZE,))
    visualize(z_arranged, color="C2", save_path="generated/z-arranged.png")


def visualize_tiled_matrix_multiplication(
    m, n, k, block_size_m, block_size_n, block_size_k
):
    BLOCK_SIZE_M = block_size_m
    BLOCK_SIZE_N = block_size_n
    BLOCK_SIZE_K = block_size_k

    a = Tensor(shape=(m, k))
    b = Tensor(shape=(k, n))
    c = Tensor(shape=(m, n))

    a_tiled = a.tile((BLOCK_SIZE_M, BLOCK_SIZE_K))
    b_tiled = b.tile((BLOCK_SIZE_K, BLOCK_SIZE_N))
    c_tiled = c.tile((BLOCK_SIZE_M, BLOCK_SIZE_N))

    a_tile = a_tiled.innermost()
    b_tile = b_tiled.innermost()
    c_tile = c_tiled.innermost()

    color_0 = "#1f77b4"
    color_1 = "#ff7f0e"
    color_2 = "#2ca02c"

    a_color = _lighten_color(color_0, 20)
    b_color = _lighten_color(color_1, 20)
    c_color = _lighten_color(color_2, 20)

    a_tile_base_color = color_0
    b_tile_base_color = color_1
    c_tile_base_color = color_2

    def _visualize_matrices(ax):
        a_verts, a_max_pos_x, a_max_pos_y = _visualize_tensor(ax, a, 0, 0, a_color)
        b_verts, b_max_pos_x, b_max_pos_y = _visualize_tensor(
            ax, b, a_max_pos_x + 2, a_max_pos_y + 2, b_color
        )
        c_verts, _, _ = _visualize_tensor(ax, c, 0, a_max_pos_y + 2, c_color)

        a_min_pos_x, a_min_pos_y = a_verts[0][0]
        b_min_pos_x, b_min_pos_y = b_verts[0][0]
        c_min_pos_x, c_min_pos_y = c_verts[0][0]

        percentage = 30

        for i in range(0, a_tiled.shape[1]):
            y_offset = i * a_tile.shape[1]

            x = a_min_pos_x + (a_tiled.shape[0] - 2) * a_tile.shape[0]
            y = a_min_pos_y + y_offset

            darkened_color = _darken_color(
                a_tile_base_color, (i + 1) / a_tiled.shape[1] * percentage
            )

            _visualize_tensor(ax, a_tile, x, y, darkened_color)

        for i in range(0, b_tiled.shape[0]):
            x_offset = i * b_tile.shape[0]

            x = b_min_pos_x + x_offset
            y = b_min_pos_y + b_tile.shape[1]

            darkened_color = _darken_color(
                b_tile_base_color,
                (b_tiled.shape[0] - i) / b_tiled.shape[0] * percentage,
            )

            _visualize_tensor(ax, b_tile, x, y, darkened_color)

        _visualize_tensor(
            ax,
            c_tile,
            c_min_pos_x + (a_tiled.shape[0] - 2) * a_tile.shape[0],
            c_min_pos_y + b_tile.shape[1],
            _darken_color(c_tile_base_color, percentage),
        )

        return b_max_pos_x, b_max_pos_y

    max_pos_x, max_pos_y = _visualize_matrices(plt.gca())

    width = max_pos_y + 1
    height = max_pos_x + 1

    _, ax = _prepare_figure_and_axes(width, height)

    _visualize_matrices(ax)

    save_path = "generated/tiled-matrix-multiplication.png"
    plt.savefig(save_path, transparent=True, bbox_inches="tight", pad_inches=0)

    plt.close()


def visualize_mm(m, n, k, block_size_m, block_size_n, block_size_k):
    BLOCK_SIZE_M = block_size_m
    BLOCK_SIZE_N = block_size_n
    BLOCK_SIZE_K = block_size_k

    input = Tensor(shape=(m, k))
    other = Tensor(shape=(k, n))
    output = Tensor(shape=(m, n))

    output_arranged = output.tile((BLOCK_SIZE_M, BLOCK_SIZE_N))
    visualize(output_arranged, color="C2", save_path="generated/output-arranged-0.png")

    input_arranged = input.tile((BLOCK_SIZE_M, BLOCK_SIZE_K))
    visualize(input_arranged, color="C0", save_path="generated/input-arranged-0.png")
    input_arranged = input_arranged.tile((1, -1))
    visualize(input_arranged, color="C0", save_path="generated/input-arranged-1.png")
    input_arranged = input_arranged.expand((-1, output_arranged.shape[1]))
    visualize(input_arranged, color="C0", save_path="generated/input-arranged-2.png")
    input_arranged.dtype = input_arranged.dtype.squeeze(0)
    visualize(input_arranged, color="C0", save_path="generated/input-arranged-3.png")

    other_arranged = other.tile((BLOCK_SIZE_K, BLOCK_SIZE_N))
    visualize(other_arranged, color="C1", save_path="generated/other-arranged-0.png")
    other_arranged = other_arranged.tile((-1, 1))
    visualize(other_arranged, color="C1", save_path="generated/other-arranged-1.png")
    other_arranged = other_arranged.expand((output_arranged.shape[0], -1))
    visualize(other_arranged, color="C1", save_path="generated/other-arranged-2.png")
    other_arranged.dtype = other_arranged.dtype.squeeze(1)
    visualize(other_arranged, color="C1", save_path="generated/other-arranged-3.png")


def _darken_color(hex_color, percentage):
    rgb = _hex_to_rgb(hex_color)
    factor = (100 - percentage) / 100
    darkened_rgb = tuple(int(c * factor) for c in rgb)

    return _rgb_to_hex(darkened_rgb)


def _lighten_color(hex_color, percentage):
    rgb = _hex_to_rgb(hex_color)
    factor = percentage / 100
    lightened_rgb = tuple(int(c + (255 - c) * factor) for c in rgb)

    return _rgb_to_hex(lightened_rgb)


def _hex_to_rgb(hex_color):
    hex_color = hex_color.lstrip("#")

    if len(hex_color) == 3:
        hex_color = "".join(c * 2 for c in hex_color)

    return tuple(int(hex_color[i : i + 2], 16) for i in (0, 2, 4))


def _rgb_to_hex(rgb_color):
    return "#{:02x}{:02x}{:02x}".format(*rgb_color)



import copy
import math
import re

import numpy as np

import ninetoothed.language
from ninetoothed.generation import CodeGenerator
from ninetoothed.symbol import Symbol
from ninetoothed.tensor import Tensor

_NUMPY = "np"


def _eval(tensor, subs=None):
    """Evaluate the symbolic tensor into a numeric tensor.

    :param tensor: The symbolic tensor.
    :param subs: The substitutions for symbolic variables.
    :return: A numeric tensor as a ``numpy.ndarray``.

    .. versionchanged:: 0.22.0

        The dimensions of the outermost level are preserved.
        Previously, the evaluation would flatten the dimensions of the
        outermost level.
    """

    if tensor.source.ndim == 0:
        return np.array(0, dtype=np.intp)

    if subs is None:
        subs = {tensor.source: {"shape": tensor.source.shape}}

    replacements = _generate_replacements(subs)

    shape = _generate_target_tensor_shape(tensor)
    shape = tuple(_replace_and_evaluate(size, replacements) for size in shape)

    if not isinstance(tensor.dtype, type(tensor)):
        return np.arange(math.prod(shape), dtype=np.intp).reshape(shape)

    result = np.empty(shape, dtype=np.intp)

    for index in np.ndindex(shape[: -tensor.innermost().ndim]):
        overall_offsets, mask = CodeGenerator._generate_overall_offsets_and_mask(
            tensor,
            tuple(Symbol(index) for index in index)
            + CodeGenerator._generate_innermost_indices(
                tensor, use_power_of_2_sizes=False
            ),
        )

        overall_offsets = _replace(str(overall_offsets), replacements)
        mask = _replace(str(mask), replacements)

        result[index + (...,)] = eval(
            f"np.where({mask}, {overall_offsets}, -1)", {_NUMPY: np}
        )

    return result


def _subs(tensor, subs):
    """Substitute symbols in the tensor.

    :param tensor: The symbolic tensor.
    :param subs: The substitutions for symbolic variables.
    :return: A new tensor with the substitutions applied.
    """

    replacements = _generate_replacements(subs)

    source = Tensor(
        shape=_replace_and_evaluate(tensor.source.shape, replacements),
        dtype=tensor.source.dtype,
        other=tensor.source.other,
        name=tensor.source.name,
    )

    def _reconstruct(source, history):
        curr = source

        for func, args, kwargs in history:
            args = _replace_and_evaluate(args, replacements)
            kwargs = _replace_and_evaluate(kwargs, replacements)

            curr = func(curr, *args, **kwargs)

        return curr

    substituted = _reconstruct(source, tensor._history)

    curr_new = substituted
    curr_old = tensor

    while isinstance(curr_new.dtype, type(tensor)):
        curr_new.dtype = _reconstruct(curr_new.dtype, curr_old.dtype._history)

        curr_new = curr_new.dtype
        curr_old = curr_old.dtype

    return substituted


def _generate_target_tensor_shape(tensor, flatten_outermost=False):
    if flatten_outermost:
        shape = [math.prod(tensor.shape)]

        curr = tensor.dtype
    else:
        shape = []

        curr = tensor

    while isinstance(curr, type(tensor)):
        shape.extend(curr.shape)

        curr = curr.dtype

    return shape


def _generate_replacements(subs):
    subs = copy.deepcopy(subs)

    replacements = {
        ninetoothed.language.LANGUAGE: _NUMPY,
        "slice(None, None, None)": ":",
    }

    for old, new in subs.items():
        if isinstance(old, Tensor):
            shape = new.shape if isinstance(new, Tensor) else new["shape"]

            for dim, size in enumerate(shape):
                replacements[old.size_string(dim)] = str(size)

            for dim, stride in enumerate(Tensor._calculate_default_strides(shape)):
                replacements[old.stride_string(dim)] = str(stride)
        elif isinstance(old, Symbol):
            replacements[str(old)] = str(new)

    return replacements


def _replace_and_evaluate(object, replacements):
    if isinstance(object, (list, tuple, set)):
        return type(object)(
            _replace_and_evaluate(item, replacements) for item in object
        )

    if isinstance(object, dict):
        return {
            key: _replace_and_evaluate(value, replacements)
            for key, value in object.items()
        }

    return eval(_replace(str(object), replacements))


def _replace(string, replacements):
    for old, new in sorted(
        replacements.items(), key=lambda key: len(key), reverse=True
    ):
        string = re.sub(rf"\b{re.escape(old)}\b", new, string)

    return string


Tensor.eval = _eval

Tensor.subs = _subs



import inspect

from ninetoothed.aot import aot
from ninetoothed.jit import jit


def make(
    arrangement,
    application,
    tensors,
    caller="torch",
    kernel_name=None,
    output_dir=None,
    num_warps=None,
    num_stages=None,
    max_num_configs=None,
):
    """Integrate the arrangement and the application of the tensors.

    :param arrangement: The arrangement of the tensors.
    :param application: The application of the tensors.
    :param tensors: The tensors.
    :param caller: Who will call the compute kernel.
    :param kernel_name: The name for the generated kernel.
    :param output_dir: The directory to store the generated files.
    :param num_warps: The number of warps to use.
    :param num_stages: The number of pipeline stages.
    :param max_num_configs: The maximum number of auto-tuning
        configurations to use.
    :return: A handle to the compute kernel.
    """

    params = inspect.signature(application).parameters
    types = arrangement(*tensors)
    annotations = {param: type for param, type in zip(params, types)}
    application.__annotations__ = annotations

    if caller == "torch":
        return jit(
            application,
            caller=caller,
            kernel_name=kernel_name,
            num_warps=num_warps,
            num_stages=num_stages,
            max_num_configs=max_num_configs,
        )

    return aot(
        application,
        caller=caller,
        kernel_name=kernel_name,
        output_dir=output_dir,
        num_warps=num_warps,
        num_stages=num_stages,
    )



import ast
import inspect
import itertools
import tkinter
from tkinter import ttk

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from matplotlib.figure import Figure
from mpl_toolkits.axes_grid1 import Divider, Size

from ninetoothed.debugging import simulate_arrangement


def visualize(tensor, color=None, save_path=None):
    """Visualize a tensor as a structured grid representation.

    :param tensor: The tensor to be visualized.
    :param color: The color to be used for visualization.
    :param save_path: The path where the visualization should be saved.
    """

    if color is None:
        color = f"C{visualize.count}"

    _, max_pos_x, max_pos_y = _visualize_tensor(plt.gca(), tensor, 0, 0, color)

    plt.close()

    width = max_pos_y + 1
    height = max_pos_x + 1

    _, ax = _prepare_figure_and_axes(width, height)

    _visualize_tensor(ax, tensor, 0, 0, color)

    if save_path is not None:
        plt.savefig(save_path, transparent=True, bbox_inches="tight", pad_inches=0)

        plt.close()
    else:
        plt.show()

    visualize.count += 1


visualize.count = 0


def visualize_arrangement(arrangement, tensors):
    """Visualize the arrangement of the tensors.

    :param arrangement: The arrangement of the tensors.
    :param tensors: The tensors.
    """

    source_tensors, target_tensors = simulate_arrangement(arrangement, tensors)

    param_names = inspect.signature(arrangement).parameters.keys()
    tensor_names = tuple(param_names)[: len(source_tensors)]

    root = tkinter.Tk()
    root.title("Arrangement Simulation")

    main_frame = ttk.Frame(root)

    root.columnconfigure(0, weight=1)
    root.rowconfigure(0, weight=1)

    canvas = FigureCanvasTkAgg(Figure(figsize=(6.4, 3)), main_frame)
    control_frame = ttk.Frame(main_frame)

    main_frame.columnconfigure(0, weight=1)
    main_frame.rowconfigure(0, weight=1)

    entry_width = 10

    tensor_name_label = ttk.Label(control_frame, text="Tensor Name")
    tensor_name_combo_box = ttk.Combobox(
        control_frame, values=tensor_names, width=entry_width
    )
    source_shape_label = ttk.Label(control_frame, text="Source Shape")
    source_shape_entry = ttk.Entry(control_frame, width=entry_width)
    source_subscript_label = ttk.Label(control_frame, text="Source Subscript")
    source_subscript_entry = ttk.Entry(control_frame, width=entry_width)
    num_programs_label = ttk.Label(control_frame, text="Number of Programs")
    num_programs_entry = ttk.Entry(control_frame, width=entry_width)
    program_id_label = ttk.Label(control_frame, text="Program ID")
    program_id_entry = ttk.Entry(control_frame, width=entry_width)
    target_shape_label = ttk.Label(control_frame, text="Target Shape")
    target_shape_entry = ttk.Entry(control_frame, width=entry_width)
    target_subscript_label = ttk.Label(control_frame, text="Target Subscript")
    target_subscript_entry = ttk.Entry(control_frame, width=entry_width)

    main_frame.grid(
        column=0, row=0, sticky=(tkinter.N, tkinter.W, tkinter.E, tkinter.S)
    )

    canvas.get_tk_widget().grid(
        column=0, row=0, sticky=(tkinter.N, tkinter.W, tkinter.E, tkinter.S)
    )
    control_frame.grid(
        column=1, row=0, sticky=(tkinter.N, tkinter.W, tkinter.E, tkinter.S)
    )

    for row, (label, widget) in enumerate(
        (
            (tensor_name_label, tensor_name_combo_box),
            (source_shape_label, source_shape_entry),
            (source_subscript_label, source_subscript_entry),
            (num_programs_label, num_programs_entry),
            (program_id_label, program_id_entry),
            (target_shape_label, target_shape_entry),
            (target_subscript_label, target_subscript_entry),
        )
    ):
        label.grid(column=0, row=row, sticky=(tkinter.W, tkinter.E))
        widget.grid(column=1, row=row, sticky=(tkinter.W, tkinter.E))

    for child in itertools.chain(
        main_frame.winfo_children(), control_frame.winfo_children()
    ):
        child.grid_configure(padx=5, pady=5)

    def _update_read_only_entry(entry, text):
        entry.configure(state="active")
        entry.delete(0, tkinter.END)
        entry.insert(tkinter.END, text)
        entry.configure(state="readonly")

    def _init():
        source_tensor = source_tensors[0]
        target_tensor = target_tensors[0]

        tensor_name_combo_box.set(tensor_names[0])
        source_subscript_entry.insert(
            tkinter.END, ", ".join(("0",) * (source_tensor.ndim - 3) + ("...",))
        )
        _update_read_only_entry(num_programs_entry, str(target_tensor.shape[0]))
        program_id_entry.insert(tkinter.END, "0")
        target_subscript_entry.insert(
            tkinter.END, ", ".join(("0",) * (target_tensor.ndim - 4) + ("...",))
        )

    _init()

    def _visualize(*_):
        canvas.figure.clear()

        tensor_name = tensor_name_combo_box.get()
        source_subscript = ast.literal_eval(source_subscript_entry.get())
        program_id = ast.literal_eval(program_id_entry.get())
        target_subscript = ast.literal_eval(target_subscript_entry.get())

        if not isinstance(source_subscript, tuple):
            source_subscript = (source_subscript,)

        if not isinstance(target_subscript, tuple):
            target_subscript = (target_subscript,)

        tensor_id = tensor_names.index(tensor_name)

        source_tensor = source_tensors[tensor_id]
        target_tensor = target_tensors[tensor_id]

        _update_read_only_entry(source_shape_entry, str(tuple(source_tensor.shape)))
        _update_read_only_entry(target_shape_entry, str(tuple(target_tensor.shape[1:])))

        source_ax, target_ax = canvas.figure.subplots(
            1, 2, subplot_kw=dict(projection="3d", aspect="equal")
        )

        _visualize_mapping(
            source_ax,
            target_ax,
            source_tensor[source_subscript],
            target_tensor[(program_id, *target_subscript)],
        )

        for ax in (source_ax, target_ax):
            ax.axis(False)

        canvas.draw()

    root.bind("<Return>", _visualize)
    tensor_name_combo_box.bind("<<ComboboxSelected>>", _visualize)

    _visualize()

    root.mainloop()


def _prepare_figure_and_axes(width, height):
    outline_width = 0.1
    plt.rcParams["lines.linewidth"] = 72 * outline_width

    fig = plt.figure(figsize=(width + outline_width, height + outline_width))

    h = (Size.Fixed(0), Size.Fixed(width + outline_width))
    v = (Size.Fixed(0), Size.Fixed(height + outline_width))

    divider = Divider(fig, (0, 0, 1, 1), h, v, aspect=False)

    ax = fig.add_axes(
        divider.get_position(), axes_locator=divider.new_locator(nx=1, ny=1)
    )

    ax.set_aspect("equal")
    ax.invert_yaxis()

    plt.axis("off")

    half_outline_width = outline_width / 2
    plt.xlim((-half_outline_width, width + half_outline_width))
    plt.ylim((-half_outline_width, height + half_outline_width))

    return fig, ax


def _visualize_tensor(ax, tensor, x, y, color, level_spacing=4):
    verts = _visualize_level(ax, tensor, x, y, color)

    if tensor.dtype is None:
        return verts, verts[1][1][0], verts[1][1][1]

    next_x, next_y = verts[0][1]
    next_y += level_spacing + 1

    next_verts, max_pos_x, max_pos_y = _visualize_tensor(
        ax, tensor.dtype, next_x, next_y, color
    )

    conn_verts = _verts_of_rect(1, level_spacing, next_x, next_y - level_spacing)
    conn_verts = [list(vert) for vert in conn_verts]
    conn_verts[2][0] += next_verts[1][0][0]

    pos_y, pos_x = zip(*conn_verts)
    pos_x = pos_x + (pos_x[0],)
    pos_y = pos_y + (pos_y[0],)

    ax.plot(pos_x[1:3], pos_y[1:3], "k--")
    ax.plot(pos_x[3:5], pos_y[3:5], "k--")

    max_pos_x = max(max_pos_x, verts[1][1][0])
    max_pos_y = max(max_pos_y, verts[1][1][1])

    return verts, max_pos_x, max_pos_y


def _visualize_level(ax, level, x, y, color):
    offsets = [1 for _ in range(level.ndim)]

    for dim in range(-3, -level.ndim - 1, -1):
        offsets[dim] = offsets[dim + 2] * level.shape[dim + 2] + 1

    indices = np.indices(level.shape)
    flattened_indices = np.stack(
        [indices[i].flatten() for i in range(level.ndim)], axis=-1
    )

    max_pos_x = x
    max_pos_y = y

    for indices in flattened_indices:
        pos = [x, y]

        for dim, index in enumerate(indices):
            pos[(level.ndim - dim) % 2] += index * offsets[dim]

        max_pos_x = max(max_pos_x, pos[0])
        max_pos_y = max(max_pos_y, pos[1])

        _visualize_unit_square(ax, pos[1], pos[0], color)

    verts = (((x, y), (x, max_pos_y)), ((max_pos_x, y), (max_pos_x, max_pos_y)))

    return verts


def _visualize_unit_square(ax, x, y, color):
    _visualize_rect(ax, 1, 1, x, y, color)


def _visualize_rect(ax, width, height, x, y, color):
    ax.add_patch(
        plt.Rectangle(
            (x, y),
            width,
            height,
            edgecolor="k",
            facecolor=color,
            linewidth=plt.rcParams["lines.linewidth"],
        )
    )


def _verts_of_rect(width, height, x, y):
    return ((x, y), (x + width, y), (x + width, y + height), (x, y + height))


def _visualize_mapping(source_ax, target_ax, source, target):
    def _to_3d_array(tensor):
        ndim = tensor.ndim

        if ndim > 3:
            raise ValueError("Tensor must have 3 or fewer dimensions.")

        shape = tuple(1 for _ in range(3 - ndim)) + tensor.shape

        return tensor.reshape(shape).permute(2, 1, 0).numpy(force=True)

    def _visualize_voxels(ax, filled, colors):
        def _explode(data):
            shape = np.array(data.shape) * 2 - 1

            if data.ndim == 4:
                shape[-1] = data.shape[-1]

            exploded = np.zeros(shape, dtype=data.dtype)
            exploded[::2, ::2, ::2, ...] = data

            return exploded

        filled = _explode(filled)
        colors = _explode(colors)

        ax.set_proj_type("ortho")
        ax.view_init(elev=-75, azim=-45, roll=-45)

        x, y, z = np.indices(np.array(filled.shape) + 1).astype(float) // 2
        x[0::2, :, :] += 0.001
        y[:, 0::2, :] += 0.001
        z[:, :, 0::2] += 0.001
        x[1::2, :, :] += 0.999
        y[:, 1::2, :] += 0.999
        z[:, :, 1::2] += 0.999

        ax.voxels(
            x, y, z, filled, facecolors=colors, edgecolor=(0, 0, 0, alpha), shade=False
        )

        ax.axis("equal")

    source = _to_3d_array(source)
    target = _to_3d_array(target)

    def _normalize(tensor, epsilon=1e-6):
        target_min = target[target >= 0].min()
        target_max = target[target >= 0].max()

        return (tensor - target_min) / (target_max - target_min + epsilon)

    color_map = mpl.colormaps["viridis"]

    alpha = 0.5

    other_color = (1, 1, 1, alpha)

    source_colors = np.where(
        np.isin(source, target)[..., None],
        color_map(_normalize(source), alpha=alpha),
        other_color,
    )
    target_colors = np.where(
        target[..., None] >= 0, color_map(_normalize(target), alpha=alpha), other_color
    )

    _visualize_voxels(source_ax, np.full_like(source, True), source_colors)
    _visualize_voxels(target_ax, np.full_like(target, True), target_colors)



import functools

import torch
import torch.nn.functional as F

import ninetoothed
import tests.test_matmul as matmul
from ninetoothed import Tensor
from tests.skippers import skip_if_cuda_not_available


def arrangement(
    input,
    filter,
    output,
    BLOCK_SIZE_M=matmul.BLOCK_SIZE_M,
    BLOCK_SIZE_N=matmul.BLOCK_SIZE_N,
    BLOCK_SIZE_K=matmul.BLOCK_SIZE_K,
):
    input_tiled = input.tile((1, *filter.shape[1:]), strides=(-1, -1, 1, 1))
    input_squeezed = input_tiled.squeeze(1)
    input_squeezed.dtype = input_squeezed.dtype.squeeze(0)
    input_raveled = input_squeezed.ravel()
    input_flattened = input_raveled.flatten(end_dim=3).flatten(start_dim=1)

    filter_flattened = filter.flatten(start_dim=1)
    filter_permuted = filter_flattened.permute((1, 0))

    output_flattened = output.permute((0, 2, 3, 1)).flatten(end_dim=3)

    return functools.partial(
        matmul.arrangement,
        BLOCK_SIZE_M=BLOCK_SIZE_M,
        BLOCK_SIZE_N=BLOCK_SIZE_N,
        BLOCK_SIZE_K=BLOCK_SIZE_K,
    )(input_flattened, filter_permuted, output_flattened)


def conv2d(input, filter):
    n, _, h, w = input.shape
    k, _, r, s = filter.shape
    p = h - r + 1
    q = w - s + 1

    output = torch.empty((n, k, p, q), device=input.device, dtype=input.dtype)

    conv2d_kernel = ninetoothed.make(
        arrangement,
        matmul.application,
        (Tensor(4), Tensor(4, shape_options={"constexpr": True}), Tensor(4)),
        max_num_configs=50,
    )

    conv2d_kernel(input, filter, output)

    return output


@skip_if_cuda_not_available
class TestCUDA:
    @classmethod
    def setup_class(cls):
        torch.manual_seed(0)

        n, c, h, w = 4, 64, 16, 16
        k, _, r, s = 512, c, 3, 3

        cls.input = torch.randn(n, c, h, w, device="cuda")
        cls.filter = torch.randn(k, c, r, s, device="cuda")

    def test_fp16(self):
        input = type(self).input.to(torch.float16)
        filter = type(self).filter.to(torch.float16)

        assert torch.allclose(
            conv2d(input, filter), F.conv2d(input, filter), atol=0.001, rtol=0.001
        )



import functools

import pytest

import ninetoothed
import tests.test_matmul as matmul
from ninetoothed import Tensor
from tests.skippers import skip_if_cuda_not_available


@skip_if_cuda_not_available
@pytest.mark.parametrize("num_stages", (None, 3, (1, 3)))
@pytest.mark.parametrize("num_warps", (None, 4, (4, 8)))
@pytest.mark.parametrize("block_size_k", (ninetoothed.block_size(), 64))
@pytest.mark.parametrize("block_size_n", (ninetoothed.block_size(), 64))
@pytest.mark.parametrize("block_size_m", (ninetoothed.block_size(), 64))
def test_auto_tuning_generation(
    block_size_m, block_size_n, block_size_k, num_warps, num_stages
):
    auto_tuning_should_be_disabled = (
        isinstance(block_size_m, int)
        and isinstance(block_size_n, int)
        and isinstance(block_size_k, int)
        and not isinstance(num_warps, tuple)
        and not isinstance(num_stages, tuple)
    )

    arrangement = functools.partial(
        matmul.arrangement,
        BLOCK_SIZE_M=block_size_m,
        BLOCK_SIZE_N=block_size_n,
        BLOCK_SIZE_K=block_size_k,
    )

    application = matmul.application

    tensors = (Tensor(2), Tensor(2), Tensor(2))

    kernel = ninetoothed.make(
        arrangement, application, tensors, num_warps=num_warps, num_stages=num_stages
    )

    source_file = kernel._source

    with open(source_file) as f:
        contents = f.read()

        if auto_tuning_should_be_disabled:
            assert "application_with_auto_tuning" not in contents
            assert "num_warps=" in contents and "num_stages=" in contents
        else:
            assert "application_with_auto_tuning" in contents



import ast
import pathlib
import re
import subprocess
import tempfile
import uuid

import ninetoothed.naming as naming
from ninetoothed.dtype import int64
from ninetoothed.generation import CACHE_DIR, CodeGenerator
from ninetoothed.tensor import Tensor
from ninetoothed.utils import calculate_default_configs


def aot(
    func,
    caller="cuda",
    kernel_name=None,
    output_dir=None,
    num_warps=None,
    num_stages=None,
):
    default_num_warps, default_num_stages = calculate_default_configs()

    if num_warps is None:
        num_warps = default_num_warps

    if num_stages is None:
        num_stages = default_num_stages

    output_dir = pathlib.Path(output_dir)

    output_contents = _aot(func, caller, kernel_name, num_warps, num_stages)

    for output_name, output_content in output_contents.items():
        output_path = output_dir / f"{kernel_name}{output_name[-2:]}"

        with open(output_path, "w") as f:
            f.write(output_content)


def _aot(func, caller, kernel_name, num_warps, num_stages):
    def _find_tensor_by_source_name(tensors, name):
        name = naming.remove_prefixes(name)

        for tensor in tensors:
            if naming.remove_prefixes(tensor.source.name) == name:
                return tensor

    _HEADER_PATH.parent.mkdir(exist_ok=True)

    if not _HEADER_PATH.exists() or _HEADER_PATH.read_text() != _HEADER_CONTENT:
        _HEADER_PATH.write_text(_HEADER_CONTENT)

    code_generator = CodeGenerator()
    source_file = code_generator(
        func,
        caller=caller,
        kernel_name=kernel_name,
        num_warps=num_warps,
        num_stages=num_stages,
        max_num_configs=None,
        prettify=False,
    )

    tensors = code_generator.tensors
    kernel_func = code_generator.kernel_func
    launch_func = code_generator.launch_func

    param_strings = ["stream"]
    param_types = []
    constexpr_param_indices = []

    for arg in kernel_func.args.args:
        param = arg.arg

        param_strings.append(param)

        if match := Tensor.pointer_pattern().fullmatch(param):
            source_name = match.group(1)
            tensor = _find_tensor_by_source_name(tensors, source_name)
            dtype = tensor.source.dtype

            param_types.append(f"*{dtype}")
        elif Tensor.size_pattern().fullmatch(param):
            param_types.append(int64)
        elif Tensor.stride_pattern().fullmatch(param):
            param_types.append(int64)
        else:
            source_name = param
            tensor = _find_tensor_by_source_name(tensors, source_name)
            dtype = tensor.source.dtype

            if tensor.constexpr:
                param_types.append(f"{tensor.value}")
                constexpr_param_indices.append(len(param_types) - 1)
            else:
                param_types.append(dtype)

    signature = ", ".join(param_types)

    for index in sorted(set(constexpr_param_indices), reverse=True):
        param_strings.pop(index + 1)
        param_types.pop(index)

    grid_extractor = _GridExtractor()
    launch_func = grid_extractor.visit(launch_func)
    grid_extractor.visit(code_generator.raw_grid)
    grid = f"{ast.unparse(grid_extractor.grid[0])}, 1, 1"

    signature_hash, output_contents = _compile(
        source_file, kernel_name, signature, grid, num_warps, num_stages
    )

    c_source_file_name = f"{kernel_name}.{signature_hash}.c"
    c_source_file = output_contents[c_source_file_name]

    c_header_file_name = f"{kernel_name}.{signature_hash}.h"
    c_header_file = output_contents[c_header_file_name]

    pattern = rf"\({', '.join(rf'(.*) {param}' for param in param_strings)}\)"
    c_param_type_strings = re.search(pattern, c_header_file).groups()

    unparser = _Unparser(c_param_type_strings)

    launch_func_unparsed = unparser.unparse(launch_func)
    launch_func_unparsed = launch_func_unparsed.replace(
        func.__name__, f"{kernel_name}_{signature_hash}"
    )

    c_source_file = f"{c_source_file}\n{launch_func_unparsed}\n"
    c_source_file = c_source_file.replace("<stdint.h>", f'"{_HEADER_PATH}"')
    output_contents[c_source_file_name] = c_source_file

    c_header_file = f'{c_header_file}\n#ifdef __cplusplus\nextern "C" {unparser.header};\n#else\n{unparser.header};\n#endif\n'
    c_header_file = c_header_file.replace("<stdint.h>", f'"{_HEADER_PATH}"')
    output_contents[c_header_file_name] = c_header_file

    return output_contents


_HEADER_CONTENT = """#ifndef NINETOOTHED_H
#define NINETOOTHED_H

#include <stdint.h>

typedef struct {
    void *data;
    uint64_t *shape;
    int64_t *strides;
} NineToothedTensor;

typedef void *NineToothedStream;

typedef int NineToothedResult;

#endif // NINETOOTHED_H
"""

_HEADER_PATH = CACHE_DIR / "ninetoothed.h"


class _Unparser:
    def __init__(self, param_types):
        self._param_types = param_types

    def unparse(self, node):
        method_name = "_unparse_" + node.__class__.__name__

        if hasattr(self, method_name):
            return getattr(self, method_name)(node)

        return self._generic_unparse(node)

    def _generic_unparse(self, node):
        return ast.unparse(node)

    def _unparse_Expr(self, node):
        return self.unparse(node.value)

    def _unparse_Call(self, node):
        call = ast.Call(
            func=node.func,
            args=[ast.Name(id="stream", ctx=ast.Load())]
            + [
                arg
                for arg in node.args
                if not isinstance(arg, ast.Name) or not naming.is_constexpr(arg.id)
            ],
            keywords=[],
        )

        unparsed = f"return {self._generic_unparse(call)};"

        pattern = rf"\((stream), {', '.join(r'([^,]*)' for _ in range(len(self._param_types) - 1))}\)"
        args = re.search(pattern, unparsed).groups()

        for i, (arg, type) in enumerate(zip(args, self._param_types)):
            if i != 0 and "." not in arg:
                new_arg = f"*({type} *){arg}.data"
            else:
                new_arg = f"({type}){arg}"

            unparsed = unparsed.replace(arg, new_arg)

        return unparsed

    def _unparse_FunctionDef(self, node):
        params = ["NineToothedStream stream"]
        params += [f"NineToothedTensor {arg.arg}" for arg in node.args.args]
        header = f"NineToothedResult {node.name}({', '.join(params)})"

        self.header = header

        body_lines = []

        for stmt in node.body:
            if isinstance(stmt, ast.Assign):
                continue

            stmt_unparsed = self.unparse(stmt)

            if isinstance(stmt, ast.Expr):
                stmt_unparsed = stmt_unparsed.strip()

            if not stmt_unparsed.endswith(";"):
                stmt_unparsed += ";"

            body_lines.append("    " + stmt_unparsed)

        body = "\n".join(body_lines)

        return f"{header} {{\n{body}\n}}"


class _GridExtractor(ast.NodeTransformer):
    def visit_BinOp(self, node):
        self.generic_visit(node)

        if isinstance(node.op, ast.FloorDiv):
            node.op = ast.Div()

        return node

    def visit_Call(self, node):
        self.generic_visit(node)

        node.func = node.func.value

        return node

    def visit_Lambda(self, node):
        self.generic_visit(node)

        self.grid = node.body.elts

        return node


def _compile(path, name, signature, grid, num_warps, num_stages):
    with tempfile.TemporaryDirectory() as temp_dir:
        output_dir = pathlib.Path(temp_dir)
        output_name = uuid.uuid4().hex
        output_path = output_dir / output_name

        command = [
            "python",
            "-m",
            "triton.tools.compile",
            str(path),
            "--kernel-name",
            str(name),
            "--signature",
            str(signature),
            "--grid",
            str(grid),
            "--num-warps",
            str(num_warps),
            "--num-stages",
            str(num_stages),
            "--out-path",
            str(output_path),
        ]

        subprocess.run(command, check=True)

        matching_files = list(output_dir.glob(f"{output_name}.*"))

        signature_hash = matching_files[0].name.split(".")[1]

        output_contents = {}

        for file in matching_files:
            with file.open() as f:
                output_contents[file.name.replace(output_name, name)] = f.read()

    return signature_hash, output_contents



import triton


def calculate_default_configs():
    device = triton.runtime.driver.active.get_current_device()
    properties = triton.runtime.driver.active.utils.get_device_properties(device)
    max_shared_mem = properties["max_shared_mem"]

    num_warps = 8
    num_stages = max_shared_mem // 2**15

    return num_warps, num_stages



import random

import torch

import ninetoothed
import ninetoothed.language as ntl
from ninetoothed import Tensor, block_size
from tests.skippers import skip_if_cuda_not_available


def arrangement(input, p, seed, output, BLOCK_SIZE=block_size()):
    return input.tile((BLOCK_SIZE,)), p, seed, output.tile((BLOCK_SIZE,))


def application(input, p, seed, output):
    output = ntl.where(ntl.rand(seed, input.offsets()) > p, input / (1 - p), 0)  # noqa: F841


def dropout(input, p=0.5):
    seed = random.randrange(0, 2**31)
    output = torch.empty_like(input)

    tensors = (Tensor(1), Tensor(0), Tensor(0), Tensor(1))
    dropout_kernel = ninetoothed.make(arrangement, application, tensors)

    dropout_kernel(input, p, seed, output)

    return output


@skip_if_cuda_not_available
class TestCUDA:
    @classmethod
    def setup_class(cls):
        random.seed(0)
        torch.manual_seed(0)

        size = 349

        cls.input = torch.randn(size, device="cuda")

    def test_fp16(self):
        input = type(self).input.to(torch.float16)
        p = 0.3

        output = dropout(input, p=p)

        assert input.shape == output.shape

        non_zero_ratio = output.nonzero().numel() / input.numel()

        assert abs(non_zero_ratio - (1 - p)) < 0.05

        assert torch.allclose(output[output != 0], input[output != 0] / (1 - p))



import importlib
import sys

from ninetoothed.generation import CodeGenerator
from ninetoothed.utils import calculate_default_configs


def jit(
    func=None,
    *,
    caller="torch",
    kernel_name=None,
    num_warps=None,
    num_stages=None,
    max_num_configs=None,
    _prettify=False,
):
    """A decorator for generating compute kernels.

    :param func: The function to be compiled.
    :param caller: Who will call the compute kernel.
    :param kernel_name: The name for the generated kernel.
    :param num_warps: The number of warps to use.
    :param num_stages: The number of pipeline stages.
    :param max_num_configs: The maximum number of auto-tuning
        configurations to use.
    :param _prettify: Whether to prettify the generated code.
    :return: A handle to the compute kernel.

    .. note::

        The ``_prettify`` parameter is experimental, which might break
        the generated code.
    """

    default_num_warps, default_num_stages = calculate_default_configs()

    if num_warps is None:
        num_warps = default_num_warps

    if num_stages is None:
        num_stages = default_num_stages

    def wrapper(func):
        return JIT(
            func,
            caller=caller,
            kernel_name=kernel_name,
            num_warps=num_warps,
            num_stages=num_stages,
            max_num_configs=max_num_configs,
            _prettify=_prettify,
        )()

    if func is None:
        return wrapper

    return wrapper(func)


class JIT:
    def __init__(
        self,
        func,
        caller,
        kernel_name,
        num_warps,
        num_stages,
        max_num_configs,
        _prettify=False,
    ):
        self.func = func

        self._caller = caller

        if kernel_name is not None:
            self._kernel_name = kernel_name
        else:
            self._kernel_name = func.__name__

        self._num_warps = num_warps

        self._num_stages = num_stages

        self._max_num_configs = max_num_configs

        self._prettify = _prettify

    def __call__(self):
        code_generator = CodeGenerator()
        source_file = code_generator(
            self.func,
            self._caller,
            self._kernel_name,
            self._num_warps,
            self._num_stages,
            self._max_num_configs,
            self._prettify,
        )
        module = import_from_path(source_file, source_file)
        module_vars = vars(module)

        handle = _Handle(
            module_vars[self._kernel_name],
            module_vars[code_generator.launch_func_name],
            source_file,
        )

        return handle


def import_from_path(module_name, file_path):
    spec = importlib.util.spec_from_file_location(module_name, file_path)
    module = importlib.util.module_from_spec(spec)
    sys.modules[module_name] = module
    spec.loader.exec_module(module)

    return module


class _Handle:
    def __init__(self, kernel, launch, source):
        self._kernel = kernel
        self._launch = launch
        self._source = source

    def __call__(self, *args, **kwargs):
        return self._launch(*args, **kwargs)



import ast
import inspect
import numbers
import types

import ninetoothed.naming as naming


class Symbol:
    """A class uesed to represent a symbol.

    :param expr: The expression used to construct the symbol.
    :param constexpr: Whether the symbol is a constexpr.
    :param mata: Whether the symbol is a meta.
    :param lower_bound: The minimum value for the symbol's range.
    :param upper_bound: The maximum value for the symbol's range.
    :param power_of_two: Whether the value should be a power of two.
    """

    def __init__(
        self,
        expr,
        constexpr=None,
        meta=None,
        lower_bound=None,
        upper_bound=None,
        power_of_two=None,
    ):
        if isinstance(expr, type(self)):
            self._node = expr._node
            return

        if isinstance(expr, ast.AST):
            self._node = expr
            return

        if isinstance(expr, types.CodeType):
            expr = inspect.getsource(expr)

        if not isinstance(expr, str):
            expr = str(expr)

        self._node = ast.parse(expr, mode="eval").body

        if (constexpr or meta) and not isinstance(self._node, ast.Name):
            raise ValueError("`constexpr` and `meta` are properties of name symbols.")

        if meta:
            if constexpr is False:
                raise ValueError("Non-constexpr meta symbol is not supported.")

            self._node.id = naming.make_meta(self._node.id)

        if constexpr:
            self._node.id = naming.make_constexpr(self._node.id)

        self._node.symbol = self

        DEFAULT_LOWER_BOUND_FOR_META_SYMBOLS = 2**5
        DEFAULT_UPPER_BOUND_FOR_META_SYMBOLS = 2**10
        DEFAULT_POWER_OF_TWO_FOR_META_SYMBOLS = True

        DEFAULT_LOWER_BOUND_FOR_NON_META_CONSTEXPR_SYMBOLS = 1
        DEFAULT_UPPER_BOUND_FOR_NON_META_CONSTEXPR_SYMBOLS = 2**20
        DEFAULT_POWER_OF_TWO_FOR_NON_META_CONSTEXPR_SYMBOLS = False

        if lower_bound is not None:
            self.lower_bound = lower_bound
        else:
            if meta:
                self.lower_bound = DEFAULT_LOWER_BOUND_FOR_META_SYMBOLS
            elif constexpr:
                self.lower_bound = DEFAULT_LOWER_BOUND_FOR_NON_META_CONSTEXPR_SYMBOLS

        if upper_bound is not None:
            self.upper_bound = upper_bound
        else:
            if meta:
                self.upper_bound = DEFAULT_UPPER_BOUND_FOR_META_SYMBOLS
            elif constexpr:
                self.upper_bound = DEFAULT_UPPER_BOUND_FOR_NON_META_CONSTEXPR_SYMBOLS

        if power_of_two is not None:
            self.power_of_two = power_of_two
        else:
            if meta:
                self.power_of_two = DEFAULT_POWER_OF_TWO_FOR_META_SYMBOLS
            elif constexpr:
                self.power_of_two = DEFAULT_POWER_OF_TWO_FOR_NON_META_CONSTEXPR_SYMBOLS

    def __eq__(self, other):
        if isinstance(self._node, ast.Constant):
            if isinstance(other, Symbol) and isinstance(other._node, ast.Constant):
                return self._node.value == other._node.value

            if isinstance(other, numbers.Number):
                return self._node.value == other

        if isinstance(self._node, ast.Name):
            if isinstance(other, Symbol) and isinstance(other._node, ast.Name):
                return (
                    self._node.id == other._node.id
                    and self._node.ctx == other._node.ctx
                )

        return False

    def __hash__(self):
        return hash(str(self))

    def __add__(self, other):
        other = type(self)(other)

        if self == 0:
            return other

        if other == 0:
            return self

        return type(self)(ast.BinOp(left=self._node, op=ast.Add(), right=other._node))

    def __radd__(self, other):
        return self.__add__(other)

    def __sub__(self, other):
        other = type(self)(other)

        if self == 0:
            return -other

        if other == 0:
            return self

        return type(self)(ast.BinOp(left=self._node, op=ast.Sub(), right=other._node))

    def __rsub__(self, other):
        return type(self)(other).__sub__(self)

    def __mul__(self, other):
        other = type(self)(other)

        if self == 0 or other == 0:
            return type(self)(0)

        if self == 1:
            return other

        if other == 1:
            return self

        return type(self)(ast.BinOp(left=self._node, op=ast.Mult(), right=other._node))

    def __rmul__(self, other):
        return self.__mul__(other)

    def __floordiv__(self, other):
        other = type(self)(other)

        if isinstance(other._node, ast.Constant) and other._node.value == 1:
            return self

        return type(self)(
            ast.BinOp(left=self._node, op=ast.FloorDiv(), right=other._node)
        )

    def __mod__(self, other):
        other = type(self)(other)

        return type(self)(ast.BinOp(left=self._node, op=ast.Mod(), right=other._node))

    def __lt__(self, other):
        other = type(self)(other)

        return type(self)(
            ast.Compare(left=self._node, ops=[ast.Lt()], comparators=[other._node])
        )

    def __and__(self, other):
        other = type(self)(other)

        return type(self)(
            ast.BinOp(left=self._node, op=ast.BitAnd(), right=other._node)
        )

    def __rand__(self, other):
        return self.__and__(other)

    def __getitem__(self, key):
        return type(self)(ast.Subscript(value=self._node, slice=type(self)(key)._node))

    def __repr__(self):
        return ast.unparse(self._node)

    def find_and_replace(self, target, replacement):
        if isinstance(target, tuple):
            targets = tuple(item.node for item in target)
        else:
            targets = (target.node,)

        return Symbol(_FindAndReplacer(targets, replacement.node).visit(self._node))

    def names(self):
        class NameCollector(ast.NodeVisitor):
            def __init__(self):
                self.names = set()

            def visit_Name(self, node):
                self.generic_visit(node)

                self.names.add(node.symbol)

        name_collector = NameCollector()

        name_collector.visit(self._node)

        return name_collector.names

    @property
    def node(self):
        class SliceSimplifier(ast.NodeTransformer):
            def visit_Call(self, node):
                if isinstance(node.func, ast.Name) and node.func.id == "slice":
                    return ast.Slice(*node.args)

                return node

        return SliceSimplifier().visit(self._node)

    @staticmethod
    def is_name(object):
        return isinstance(object, Symbol) and isinstance(object.node, ast.Name)


def block_size(lower_bound=None, upper_bound=None):
    """Create a block size symbol that serves as a meta-parameter.

    :param lower_bound: The lower bound for the block size's range.
    :param upper_bound: The upper bound for the block size's range.
    :return: A block size symbol that serves as a meta-parameter.
    """

    name = naming.auto_generate(f"BLOCK_SIZE_{block_size._num_block_sizes}")

    block_size._num_block_sizes += 1

    return Symbol(name, meta=True, lower_bound=lower_bound, upper_bound=upper_bound)


block_size._num_block_sizes = 0


class _FindAndReplacer(ast.NodeTransformer):
    def __init__(self, targets, replacement):
        self._targets_unparsed = tuple(
            sorted({ast.unparse(target) for target in targets}, key=len, reverse=True)
        )
        self._replacement = replacement

    def visit(self, node):
        if ast.unparse(node) in self._targets_unparsed:
            return self._replacement

        return super().visit(node)
